/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * Author: Charles Kerr <charles@rebelbase.com>
 *
 * Copyright (C) 2002  Charles Kerr <charles@rebelbase.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * A note on implementation
 * The queue defers all public tasks (like insert, remove, etc.)
 * to the queue thread.  This ensures that all tasks are handled
 * in sequence, rather than concurrently.  This ensures that callbacks
 * will be invoked in sync with the queue, so the queue's state won't
 * change during the callback.
 */

/*********************
**********************  Includes
*********************/

#include <config.h>

#include <errno.h>

#include <glib.h>

#include <unistd.h>

#include <pan/base/debug.h>
#include <pan/base/log.h>
#include <pan/base/pan-glib-extensions.h>
#include <pan/base/pan-i18n.h>
#include <pan/base/server.h>

#include <pan/gui.h>
#include <pan/prefs.h>
#include <pan/nntp.h>
#include <pan/queue.h>
#include <pan/server-ui.h>

/*********************
**********************  Defines / Enumerated types
*********************/

typedef enum
{
	TODO_INSERT,
	TODO_REMOVE,
	TODO_MOVE,
	TODO_REQUEUE,
	TODO_TASK_ABORT,
	TODO_SET_MAX_SOCKET_QTY,
	TODO_SET_MAX_TRIES,
	TODO_SHUTDOWN,
	TODO_ONLINE
}
QueueTodoAction;

/*********************
**********************  Macros
*********************/

/*********************
**********************  Structures / Typedefs
*********************/

typedef struct
{
	GSList * tasks;
	int index_1;
	int index_2;
	QueueTodoAction action;
}
QueueTodo;

typedef struct
{
	PanSocket *socket;
	Server *server;
	Task *task;	/* NULL if idle */
}
SocketInfo;

/*********************
**********************  Private Function Prototypes
*********************/

static void fire_online_status_changed (gboolean online);

static void* queue_mainloop (void*);

static void queue_run_what_we_can (void);

static void queue_set_task_status (Task* task,
                                   QueueTaskStatus status);

static void socket_cleanup (Task* task);

static void queue_new_todo (QueueTodoAction action,
                            GSList * tasks,
                            int i1,
                            int i2);

static guint queue_get_length (void);

static void queue_do_todo (void);

/*********************
**********************  Constants
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

/***********
************  Private
***********/

/*********************
**********************  Variables
*********************/

/***********
************  Extern
***********/

/***********
************  Public
***********/

PanCallback * queue_task_added           = NULL;
PanCallback * queue_tasks_removed         = NULL;
PanCallback * queue_tasks_moved           = NULL;
PanCallback * queue_task_status_changed  = NULL;
PanCallback * queue_max_tries_changed    = NULL;

/***********
************  Private
***********/


/* FIXME glib 2.0: for efficiency of adding to end, use a GQueue instead of GSList */
static GQueue * todo_queue = NULL;
static GSList* task_list = NULL;

static GMutex * sock_lock = NULL;
static GMutex * task_lock = NULL;
static GMutex * cond_mutex = NULL;
static GMutex * todo_mutex = NULL;

static GCond * qcond = NULL;
static gboolean work_to_do = FALSE;

static GHashTable* task_to_status = NULL;

static GSList * sockets = NULL;

static int max_socket_qty = 4;

static int max_tries = 5;

static guint running_tasks = 0;

static gboolean _online = TRUE;

/*********************
**********************  BEGINNING OF SOURCE
*********************/

/************
*************  PUBLIC ROUTINES
************/

void
queue_wakeup (void)
{
	debug_enter ("queue_wakeup");
        g_mutex_lock (cond_mutex);
	work_to_do = TRUE;
        g_cond_signal (qcond);
        g_mutex_unlock (cond_mutex);
	debug_exit ("queue_wakeup");
}

void
queue_init (int socket_qty, int tries)
{
	debug_enter ("queue_init");

	task_to_status = g_hash_table_new (g_direct_hash, g_direct_equal);

	todo_queue = g_queue_new ();

	queue_task_added = pan_callback_new ();
	queue_tasks_removed = pan_callback_new ();
	queue_tasks_moved = pan_callback_new ();
	queue_task_status_changed = pan_callback_new ();
	queue_max_tries_changed = pan_callback_new ();

	sock_lock = g_mutex_new ();
	task_lock = g_mutex_new ();
	cond_mutex = g_mutex_new ();
	todo_mutex = g_mutex_new ();

	qcond = g_cond_new ();

	max_tries = tries;
	max_socket_qty = socket_qty;

	/* fire up the queue thread */
	g_thread_create (queue_mainloop, NULL, FALSE, NULL);

	debug_exit ("queue_init");
}


gboolean
queue_is_online (void)
{
	return _online;
}
void
queue_set_online (gboolean online)
{
	queue_new_todo (TODO_ONLINE, NULL, online, -1);
}
static void
real_queue_set_online (gint p)
{
	if (_online == p)
	{
		if (_online)
			queue_wakeup ();
	}
	else
	{
		_online = p;

		fire_online_status_changed (_online);

		if (_online)
			queue_wakeup ();
		else
		{
			GSList * l;

			g_mutex_lock (sock_lock);
			for (l=sockets; l!=NULL; l=l->next)
			{
				SocketInfo * i = (SocketInfo*)l->data;
				task_hint_abort (i->task);
				queue_set_task_status (i->task, QUEUE_TASK_STATUS_ABORTING);
			}
			g_mutex_unlock (sock_lock);
		}
	}
}


void
queue_shutdown (void)
{
	queue_new_todo (TODO_SHUTDOWN, NULL, 0, -1);
}
static void
real_queue_shutdown (void)
{
	gint len;
	GSList * l;
	GSList * sockets_tmp;

	/* clear the status hashtable */
	g_hash_table_destroy (task_to_status);
	task_to_status = NULL;



	/* optionally save the tasks */

	/* unref the tasks */

	/* get the socket array */
	g_mutex_lock (sock_lock);
	sockets_tmp = sockets;
	sockets = NULL;
	g_mutex_unlock (sock_lock);

	/* close the sockets */
	len = g_slist_length (sockets_tmp);
	for (l=sockets_tmp; l!=NULL; l=l->next)
		pan_object_unref (PAN_OBJECT(((SocketInfo*)l->data)->socket));
	gui_inc_connection_size (-len);
	g_slist_foreach (sockets_tmp, (GFunc)g_free, NULL);
	g_slist_free (sockets_tmp);
}


void
queue_add (Task * task)
{
	g_return_if_fail (task!=NULL);
	queue_task_insert (task, task->high_priority ? 0 : -1);
}

GPtrArray*
queue_get_tasks (void)
{
	GSList * l;
	GPtrArray * a = g_ptr_array_new ();

        g_mutex_lock (task_lock);
	for (l=task_list; l!=NULL; l=l->next) {
		pan_object_ref (PAN_OBJECT(l->data));
		g_ptr_array_add (a, l->data);
	}
        g_mutex_unlock (task_lock);

	return a;
}

/**
***
**/

int
queue_get_max_sockets (void)
{
	return max_socket_qty;
}
void
queue_set_max_sockets (int i)
{
	queue_new_todo (TODO_SET_MAX_SOCKET_QTY, NULL, i, 0);
	queue_wakeup ();
}
static void
real_queue_set_max_sockets (int i)
{
	max_socket_qty = i;
	/* FIXME: fire some callback? */
}

guint
queue_get_running_task_count (void)
{
	gint running_count = 0;
	GSList *l;

	g_mutex_lock (task_lock);
	for (l=task_list; l!=NULL; l=l->next)
	{
		QueueTaskStatus status = queue_get_task_status (l->data);

		if (status == QUEUE_TASK_STATUS_RUNNING ||
		    status == QUEUE_TASK_STATUS_QUEUED ||
		    status == QUEUE_TASK_STATUS_CONNECTING)
			++running_count;
	}
	g_mutex_unlock (task_lock);

	return running_count;
}

/**
***
**/

void
queue_abort_tasks (GSList * tasks)
{
	queue_new_todo (TODO_TASK_ABORT, tasks, -1, -1);
}
static void
real_queue_abort_tasks (GSList * tasks)
{
	GSList * l;
	for (l=tasks; l!=NULL; l=l->next)
	{
		Task * task = TASK(l->data);
		QueueTaskStatus status = queue_get_task_status (task);

		if (status==QUEUE_TASK_STATUS_RUNNING)
		{
			/* ask a running task to abort */
			task_hint_abort (TASK(task));
			queue_set_task_status (task, QUEUE_TASK_STATUS_ABORTING);
		}
		else if (task->hint_abort && !queue_is_online())
		{
			/* the task's been stopped because the queue is offline */
			task->tries = 0;
			task->hint_abort = FALSE;
			queue_set_task_status (task, QUEUE_TASK_STATUS_QUEUED);
		}
		else
		{
			/* the task has failed */
			task->tries = 0;
			task->hint_abort = FALSE;
			queue_set_task_status (task, QUEUE_TASK_STATUS_FAILED);
		}
	}
}

/**
***
**/

void
queue_task_insert (Task* task, int index)
{
	if (task!=NULL)
		queue_new_todo (TODO_INSERT, g_slist_append(NULL, task), index, -1);
}

static void
real_queue_task_insert (Task* task, int index)
{
	guint task_len;
	debug2 (DEBUG_QUEUE, "queue_task_insert %p to pos %d", task, index);

	g_return_if_fail (task!=NULL);
	g_return_if_fail (index==-1 || (guint)index<=queue_get_length());

	/* add to the list */
	g_mutex_lock (task_lock);
	task_list = g_slist_insert (task_list, task, index);
	g_hash_table_insert (task_to_status, task, GINT_TO_POINTER(QUEUE_TASK_STATUS_QUEUED));
	task_len = g_slist_length (task_list);
	g_mutex_unlock (task_lock);

	gui_set_queue_size (running_tasks, task_len);

	/* call callbacks */
	debug1 (DEBUG_QUEUE, "queue size is now %u, calling callbacks", task_len);
	pan_callback_call (queue_task_added, task, GINT_TO_POINTER(index));

	queue_wakeup ();
	
	debug2 (DEBUG_QUEUE, "queue_task_insert %p to pos %d done", task, index);
}

/**
***
**/

void
queue_remove_tasks (GSList * tasks)
{
	queue_new_todo (TODO_REMOVE, tasks, -1, -1);
}
static void
real_queue_remove_tasks (GSList * tasks)
{
	GSList * l;
	GSList * removed = NULL;

	for (l=tasks; l!=NULL; l=l->next)
	{
		Task * task = TASK(l->data);

		if (task->is_running)
		{
			/* abort the task; after it's done running
			 * queue_run_thread() will send the task back here
			 * for part two  */
			task_hint_abort (task);
			queue_set_task_status (task, QUEUE_TASK_STATUS_REMOVING);
		}
		else /* not running */
		{
			/* not running */
			socket_cleanup (task);

			/* remove the task */
			g_mutex_lock (task_lock);
			task_list = g_slist_remove (task_list, task);
			removed = g_slist_prepend (removed, task);
			g_mutex_unlock (task_lock);
		}
	}

	if (removed != NULL)
	{
		removed = g_slist_reverse (removed);

		gui_set_queue_size (running_tasks, g_slist_length(task_list));
		queue_wakeup ();
		pan_callback_call (queue_tasks_removed, removed, NULL);

		for (l=removed; l!=NULL; l=l->next) {
			Task * task = TASK(l->data);
			g_hash_table_remove (task_to_status, task);
			pan_object_unref (PAN_OBJECT(task));
		}

		g_slist_free (removed);
	}
}

/**
***
**/

void
queue_move_tasks (GSList * tasks, int index)
{
	g_return_if_fail (tasks!=NULL);

	queue_new_todo (TODO_MOVE, tasks, index, -1);
}

static void
real_queue_move_tasks (GSList * tasks, int moveto_index)
{
	GSList * l;
	GSList * ref;

	/* sanity clause */
	g_return_if_fail (tasks!=NULL);
	g_return_if_fail (task_list!=NULL);

	g_mutex_lock (task_lock);

	/* find the point of reference */
	if (moveto_index == 0)
		ref = NULL;
	else {
		ref = g_slist_nth (task_list, moveto_index-1);
		while (ref != NULL) {
			if (g_slist_find (tasks, ref->data) == NULL)
				break;
			ref = ref->next;
		}
		if (ref == NULL)
			ref = g_slist_last (task_list);
	}

	/* remove the tasks */
	for (l=tasks; l!=NULL; l=l->next)
		task_list = g_slist_remove (task_list, l->data);

	/* add them back in */
	if (ref == NULL) {
		l = g_slist_copy (tasks);
		g_slist_last(l)->next = task_list;
		task_list = l;
		moveto_index = 0;
	} else {
		l = g_slist_copy (tasks);
		g_slist_last(l)->next = ref->next;
		ref->next = l;
		moveto_index = g_slist_index (task_list, tasks->data);
	}

	/* let everyone know */
	g_mutex_unlock (task_lock);
	pan_callback_call (queue_tasks_moved, tasks, GINT_TO_POINTER(moveto_index));
	debug_exit ("real_queue_move_tasks");
}

/**
***
**/

void
queue_requeue_failed_tasks (GSList * tasks)
{
	GSList * l;

	/* sanity clause */
	g_return_if_fail (tasks != NULL);
	for (l=tasks; l!=NULL; l=l->next) {
		Task * task = (Task*) l->data;
		g_return_if_fail (queue_get_task_status(task) == QUEUE_TASK_STATUS_FAILED);
	}

	queue_new_todo (TODO_REQUEUE, tasks, -1, -1);
}
static void
real_requeue_failed_tasks (GSList * tasks)
{
	GSList * l;

	for (l=tasks; l!=NULL; l=l->next) {
		Task * task = TASK(l->data);
		task->tries = 0;
		task->hint_abort = FALSE;
		queue_set_task_status (task, QUEUE_TASK_STATUS_QUEUED);
	}
}

/**
***
**/

int
queue_get_max_tries (void)
{
	return max_tries;
}

void
queue_set_max_tries (int i)
{
	queue_new_todo (TODO_SET_MAX_TRIES, NULL, i, -1);
}
static void
real_queue_set_max_tries (int i)
{
	max_tries = i;
	pan_callback_call (queue_max_tries_changed, NULL, GINT_TO_POINTER(i));
}

/**
***
**/

QueueTaskStatus
queue_get_task_status (const Task* task)
{
	QueueTaskStatus status = QUEUE_TASK_STATUS_NOT_QUEUED;

	if (task_to_status != NULL) {
		gpointer p = g_hash_table_lookup (task_to_status, task);
		if (p != NULL)
			status = GPOINTER_TO_INT (p);
	}

	return status;
}

static void
queue_set_task_status (Task* task, QueueTaskStatus status)
{
	g_return_if_fail (task!=NULL);
	g_return_if_fail (g_slist_index(task_list, task) != -1);

	g_hash_table_insert (task_to_status, task,
	                     GINT_TO_POINTER(status));
	pan_callback_call (queue_task_status_changed, task,
	                   GINT_TO_POINTER(status)); 
}

/************
*************  PRIVATE ROUTINES
************/

static void
queue_new_todo (QueueTodoAction action, GSList * tasks, int i1, int i2)
{
	QueueTodo *todo = g_new (QueueTodo, 1);
	todo->tasks = tasks;
	todo->index_1 = i1;
	todo->index_2 = i2;
	todo->action = action;

	g_mutex_lock (todo_mutex);
	g_queue_push_tail (todo_queue, todo);
	debug3 (DEBUG_QUEUE, "todo queue now has %u tasks (new task type: %d, int 1: %d)",
		todo_queue->length, action, i1);
	g_mutex_unlock (todo_mutex);

	queue_wakeup ();
}

static guint
queue_get_length (void)
{
	guint size;
	g_mutex_lock (task_lock);
	size = g_slist_length (task_list);
	g_mutex_unlock (task_lock);
	return size;
}

static void*
queue_run_thread (void* vp)
{
	int status;
	int result;
	gboolean aborting;
	Task* task = TASK(vp);
	gchar* desc = status_item_describe (STATUS_ITEM(task));
	debug2 (DEBUG_QUEUE, "queue_run_thread: task %p [%s]", task, desc);

	/* run */
	status_item_set_active (STATUS_ITEM(task), TRUE);
	if (task->sock)
		pan_socket_reset_statistics (task->sock);
	gui_set_queue_size (++running_tasks, queue_get_length());
	result = task_run (task);
	gui_set_queue_size (--running_tasks, queue_get_length());
	debug3 (DEBUG_QUEUE, "queue task[%p] (%s) returned: %d", task, desc, result);
	g_free(desc);

	/* clean up gui */
	status_item_set_active (STATUS_ITEM(task), FALSE);

	/* clean up socket.
	   We throw away any socket of a failed or aborting task,
	   just in case the read buffer has junk in there. */
	status = queue_get_task_status (task);
	aborting = status==QUEUE_TASK_STATUS_ABORTING || status==QUEUE_TASK_STATUS_REMOVING;
	if (task->sock!=NULL && (aborting || result!=TASK_SUCCESS))
		task->sock->error = TRUE;
	socket_cleanup (task);

	/* clean up the task */
	task->is_running = FALSE;

	if (result==TASK_SUCCESS) 
	{
		queue_remove_tasks (g_slist_append(NULL,task));
	}
	else if (status == QUEUE_TASK_STATUS_REMOVING)
	{
		queue_remove_tasks (g_slist_append(NULL,task));
	}
	else if (result==TASK_FAIL_HOPELESS && remove_failed_tasks) 
	{
		queue_remove_tasks (g_slist_append(NULL,task));
	}
	else if (result==TASK_FAIL && task->tries<max_tries && !aborting) /* retry */
	{
		++task->tries;
		queue_set_task_status (task, QUEUE_TASK_STATUS_QUEUED);
		queue_wakeup ();
	}
	else if (status == QUEUE_TASK_STATUS_ABORTING)
	{
		task->tries = 0;
		queue_abort_tasks (g_slist_append(NULL,task));
	}
	else /* failed */
	{
		task->tries = 0;
		task->hint_abort = FALSE;
		queue_set_task_status (task, QUEUE_TASK_STATUS_FAILED);
		queue_wakeup ();
	}

	debug1 (DEBUG_QUEUE, "queue_run_thread: task %p done", vp);
	g_thread_exit (NULL);
	return NULL;
}


static void
queue_run (Task* task)
{
	g_return_if_fail (task!=NULL);
	debug2 (DEBUG_QUEUE, "starting a thread to run task %p, sock %p", task, task->sock);

	queue_set_task_status (task, QUEUE_TASK_STATUS_RUNNING);

	task->is_running = TRUE;
	g_thread_create (queue_run_thread, task, FALSE, NULL);
}


/**
 * Allocates a socket for a task.  First it looks for an idle socket in the
 * socket pool; otherwise, if we're not maxed out, it tries to create a new
 * socket.
 *
 * Return -1 if we can't connect because all lines are busy
 * Return -2 if we can't connect because of a connection failure
 */
static int
queue_get_socket_for_task (Task *task)
{
	GSList* sl = NULL;
	int retval = -1;
	const gboolean gets_bodies = TASK(task)->gets_bodies;
	gboolean skip_first_idle = task->server->reserve_connection_for_bodies && !gets_bodies;
	gboolean reader_socket_exists = FALSE;
	Server * server = task->server;
	int socket_qty;
	debug_enter ("queue_get_socket_for_task");

	g_mutex_lock (sock_lock);

	/* if an idle socket for that server, use it. */
	socket_qty = 0;
	for (sl=sockets; sl!=NULL && retval!=0; sl=sl->next)
	{
		SocketInfo * si = (SocketInfo*) sl->data;

		if (si->server != server) /* ignore other servers */
			continue;

		++socket_qty;

		if (!queue_is_online()) /* offline */
			continue;
		if (si->task != NULL) /* socket already in use */
			continue;

		if (skip_first_idle) { /* save one article reader socket */
			reader_socket_exists = TRUE;
			skip_first_idle = FALSE;
			continue;
		}

		si->task = task;
		task->sock = si->socket;

		debug3 (DEBUG_QUEUE,
			"queue_get_socket_for_task %p "
			"reusing socket #%d (%p)",
			task, g_slist_index(sockets, si), si->socket);

		retval = 0;
	}

	/* if no idle sockets for that server... */
	if (retval && queue_is_online() && socket_qty<max_socket_qty)
	{
		const int current_connections = socket_qty;
		int slots_open = server->max_connections - current_connections;
		debug1 (DEBUG_QUEUE, "queue_get_socket_for_task %p no idle sockets", task);

		/* leave an article reader socket free */
		if (!reader_socket_exists
			&& server->reserve_connection_for_bodies
			&& !gets_bodies
			&& max_socket_qty>1
			&& server->max_connections>1)
				--slots_open;

		if (slots_open <= 0)
		{
			debug2 (DEBUG_QUEUE,
				"we've already got %d connections "
				"and %d reader connections",
				current_connections,
				(reader_socket_exists?1:0) );

			retval = -1;
		}
		else	
		{
			char * errmsg = NULL;

			/* open a new socket.. */
			PanSocket *sock = NULL;
			debug1 (DEBUG_QUEUE, "making a new socket to %s", server->address);

			gui_set_connecting_flag (TRUE);
			sock = pan_socket_new (server->address, server->port);
			pan_socket_set_nntp_auth (sock, server->need_auth, server->username, server->password); 

			debug1 (DEBUG_QUEUE, "new socket %p", sock);
			if (sock == NULL || sock->sockfd < 0)
			{
				if (is_nonempty_string (server->address))
					errmsg = g_strdup_printf (_("Unable to connect to server %s:%d\n"), server->address, server->port);
				else
					errmsg = g_strdup (_("No server address specified in the Preferences"));
			}
			else /* handshake */
			{
				StatusItem * status = STATUS_ITEM(task);
				gboolean posting_ok = FALSE;
				gint val = nntp_handshake (status, sock, &posting_ok);
				if (val != TASK_SUCCESS)
					errmsg = g_strdup(g_slist_last(status->errors)->data);
			}
			if (errmsg!=NULL)
			{
				log_add (LOG_ERROR, errmsg);
				g_free (errmsg);
				if (sock)
					pan_object_unref (PAN_OBJECT(sock));
				sock = NULL;

				debug1 (DEBUG_QUEUE, "%p can't log in!", sock);

				retval = -2;
			}
			else
			{
				SocketInfo * si = NULL;

				debug1 (DEBUG_QUEUE, "%p logged in", sock);

				/* looks like we got a new socket okay... */
				si = g_new (SocketInfo, 1);
				si->socket = sock;
				si->server = server;
				si->task = task;
				task->sock = sock;
				sockets = g_slist_prepend (sockets, si);
				retval = 0;
				gui_inc_connection_size (1);
			}

			gui_set_connecting_flag (FALSE);
		}
	}

	g_mutex_unlock (sock_lock);

	debug_exit ("queue_get_socket_for_item");
	return retval;
}

/**
 * Run any tasks in the task_list which can be run (ie, can get a socket,
 * if necessary) right now.
 */
static void
queue_run_what_we_can (void)
{
	GSList * l = NULL;
	debug_enter ("queue_run_what_we_can");

	/* walk through the queue and see if any tasks are waiting to be run. */
	l = task_list;
	while (l!=NULL)
	{
		Task *task = TASK(l->data);

		/* if it's not waiting to be run, leave it alone */
		if (queue_get_task_status(task) != QUEUE_TASK_STATUS_QUEUED)
		{
			debug0 (DEBUG_QUEUE, "task isn't waiting to be run");

		}

		/* if it doesn't need a socket, no problemo! */
		else if (!task->needs_socket)
		{
			debug1 (DEBUG_QUEUE,
				"task %d needs no socket; running it now",
				task);

			queue_run (task);
		}

		/* try to get a socket... */
		else if (queue_is_online() && max_socket_qty>0)
		{
			int i;
			queue_set_task_status (task, QUEUE_TASK_STATUS_CONNECTING);
			i = queue_get_socket_for_task (task);
			if (i==0)
			{
				debug1 (DEBUG_QUEUE,
					"task %d got a socket!",
					task);

				queue_run (task);
			}
			else
			{
				if (i==-2) /* connection failure */
				{
					status_item_emit_error (STATUS_ITEM(task), _("Connect Failure"));

					++task->tries;
					if (task->tries > max_tries)
						queue_abort_tasks (g_slist_append(NULL,task));
				}

				queue_set_task_status (task, QUEUE_TASK_STATUS_QUEUED);
			}
		}

		if (!g_queue_is_empty (todo_queue))
		{
				queue_do_todo ();
				l = task_list;
		}
		else
		{
				l = l->next;
		}
	}

	debug_exit ("queue_run_what_we_can done");
}

/**
 * Destroy or recycle a socket when a task is done with it.
 * @param task the finished task.
 */
static void
socket_cleanup (Task* task)
{
	debug_enter ("socket_cleanup");

	if (task && task->sock)
	{
		PanSocket* socket = task->sock;
		GSList* l = NULL;
		SocketInfo * info = NULL;
		int server_socket_qty;
		int socket_qty;

		g_mutex_lock (sock_lock);

		/* count the number of sockets this server has */
		socket_qty = 0;
		server_socket_qty = 0;
		for (l=sockets; l!=NULL; l=l->next) {
			SocketInfo * sockinfo = (SocketInfo*) l->data;
			if (sockinfo->socket == socket)
				info = sockinfo;
			if (sockinfo->server == task->server)
				++server_socket_qty;
			++socket_qty;
		}

		/* find this SocketInfo in our tables */
		g_assert (info!=NULL);

		/**
                 *  Recycle the socket if all are true
		 *  + there was no error
		 *  + the server is still online
		 *  + we don't have too many sockets for that server
		 *  + we don't have too many sockets overall
		 */
		if (!socket->error
		    && queue_is_online()
		    && socket_qty<=max_socket_qty
		    && server_socket_qty<=task->server->max_connections)
		{
		       	/* if socket's okay, reuse it */
			info->task = NULL;

			debug1 (DEBUG_QUEUE, "socket_cleanup: recycling socket %p", socket);
		}
		else
		{
			/* otherwise throw it away */
			nntp_disconnect (NULL, socket);
			pan_object_unref (PAN_OBJECT(socket));
			gui_inc_connection_size (-1);
			sockets = g_slist_remove (sockets, info);
			g_free (info);

			debug1 (DEBUG_QUEUE,
				"socket_cleanup: throwing away socket %p",
				socket);
		}
		task->sock = NULL;
		g_mutex_unlock (sock_lock);
	}
	debug_exit ("socket_cleanup");
}

static gchar*
noop_describe (const StatusItem * item)
{
	return g_strdup (_("Sending 'stay connected' request"));
}

static gchar*
disconnect_describe (const StatusItem * item)
{
	return g_strdup (_("Disconnecting"));
}

/**
 * socket upkeep:  Cull out idle sockets that have been connected too long,
 * and send keepalive commands the rest of the idle sockets so that the
 * server doesn't drop us.
 */
static void
sockets_upkeep (void)
{
	GSList* l = NULL;
	GSList* cull = NULL;
	int socket_qty;
	debug_enter ("socket_upkeep");

	g_mutex_lock (sock_lock);

	socket_qty = g_slist_length (sockets);

	for (l=sockets; l!=NULL; l=l->next)
	{
		gint age;
		gboolean destroy, idle, old, too_many, offline;
		SocketInfo * info;
		PanSocket * sock;
		const time_t current_time = time(0);

		info = (SocketInfo*) l->data;
		g_assert (info != NULL);
		sock = info->socket;
		g_assert (sock != NULL);

		/* calculate some information about the socket
		   to decide whether or not we want to keep it */
		age = current_time - sock->last_action_time;
		idle = info->task == NULL;
		old = age > info->server->idle_secs_before_timeout;
		too_many = socket_qty > max_socket_qty;
		offline = !queue_is_online ();
		destroy = idle && (old||too_many||offline);

		/* if it's idle and we're keeping it, send a keepalive msg.
		   Note that we don't let this affect the socket's
		   age, otherwise they would never grow old and close... */
		if (info->task==NULL && !destroy)
		{
			const time_t last_action_time = sock->last_action_time;
			StatusItem * status = status_item_new (noop_describe);
			status_item_set_active (status, TRUE);

			destroy = nntp_noop (STATUS_ITEM(status), sock) != TASK_SUCCESS;
			sock->last_action_time = last_action_time;

			status_item_set_active (status, FALSE);
			pan_object_unref (PAN_OBJECT(status));
		}

		/* either keep the socket or cull it from the herd */
		if (destroy) {
			--socket_qty;
			cull = g_slist_prepend (cull, info);
		}
	}

	debug1 (DEBUG_QUEUE, "sockets_upkeep: culling %u sockets",
		g_slist_length(cull));

	for (l=cull; l!=NULL; l=l->next)
	{
		const time_t current_time = time(0);
		SocketInfo * info = (SocketInfo*) l->data;
		PanSocket * sock = info->socket;
		StatusItem * status = STATUS_ITEM(status_item_new (disconnect_describe));
		const int age = current_time - sock->last_action_time;

		/* politely disconnect */
		status_item_set_active (status, TRUE);
		if (!sock->error)
			nntp_disconnect (status, sock);
		status_item_set_active (status, FALSE);
		log_add_va (LOG_INFO, _("Closing connection %p after %d seconds idle"), sock, age);

		/* destroy the socket */
		pan_object_unref (PAN_OBJECT(sock));
		gui_inc_connection_size (-1);
		sockets = g_slist_remove (sockets, info);

		/* cleanup local */
		pan_object_unref (PAN_OBJECT(status));
		g_free (info);
	}

	pan_warn_if_fail (g_slist_length(sockets) == socket_qty);

	g_mutex_unlock (sock_lock);

	g_slist_free (cull);

	debug_exit ("socket_upkeep");
}

/**
 * Process the tasks that are waiting in the todo list
 */
static void
queue_do_todo (void)
{
	gboolean do_shutdown = FALSE;
	GQueue * queue;

	g_mutex_lock (todo_mutex);
	queue = todo_queue;
	todo_queue = g_queue_new ();
	g_mutex_unlock (todo_mutex);

	while (!g_queue_is_empty(queue))
	{
		QueueTodo* a = (QueueTodo*) g_queue_pop_head (queue);
		Task * task = a->tasks ? a->tasks->data : NULL;

		switch (a->action)
		{
			case TODO_INSERT:
				real_queue_task_insert (task, a->index_1);
				break;
			case TODO_REMOVE:
				real_queue_remove_tasks (a->tasks);
				break;
			case TODO_MOVE:
				real_queue_move_tasks (a->tasks, a->index_1);
				break;
			case TODO_REQUEUE:
				real_requeue_failed_tasks (a->tasks);
				break;
			case TODO_TASK_ABORT:
				real_queue_abort_tasks (a->tasks);
				break;
			case TODO_SET_MAX_SOCKET_QTY:
				real_queue_set_max_sockets (a->index_1);
				break;
			case TODO_SET_MAX_TRIES:
				real_queue_set_max_tries (a->index_1);
				break;
			case TODO_SHUTDOWN:
				do_shutdown = TRUE;
				break;
			case TODO_ONLINE:
				real_queue_set_online (a->index_1);
				break;
			default:
				pan_warn_if_reached();
				break;
		}

		g_slist_free (a->tasks);
		g_free (a);
	}

	if (do_shutdown)
		real_queue_shutdown ();

	/* throw the todo list away */
	g_queue_free (queue);
}

static void*
queue_mainloop (void* unused)
{
	const int TIMEOUT_SECS = 60;

	for (;;)
	{
		gboolean was_timeout;
		GTimeVal sleep_until;

		g_mutex_lock (cond_mutex);

		/* sleep for TIMEOUT_SECS unless someone wakes us up */
		g_get_current_time (&sleep_until);
		g_time_val_add (&sleep_until, TIMEOUT_SECS*G_USEC_PER_SEC);
		was_timeout = FALSE;
		while (!work_to_do && !was_timeout)
			was_timeout = !g_cond_timed_wait (qcond, cond_mutex, &sleep_until);
		work_to_do = FALSE;
                g_mutex_unlock (cond_mutex);

		/* do work */
		if (was_timeout)
			sockets_upkeep ();
		else {
			queue_do_todo ();
                	queue_run_what_we_can ();
		}
	}

	return NULL;
}


PanCallback*
queue_get_online_status_changed_callback (void)
{
	static PanCallback * cb = NULL;
	if (cb==NULL) cb = pan_callback_new ();
	return cb;
}

static void
fire_online_status_changed (gboolean online)
{
	pan_callback_call (queue_get_online_status_changed_callback(),
	                   GINT_TO_POINTER(online), NULL);
}
